
import io
import pathlib
import re
from functools import cached_property, partial
from typing import (
    AsyncIterator, Dict, Generator, Iterator, List, Optional,
    Pattern, Set, Tuple)

from yamllint import linter  # type:ignore
from yamllint.config import YamlLintConfig  # type:ignore

import abstracts

from aio.core.dev import debug
from aio.core import directory
from aio.core.functional import (
    async_property,
    AwaitableGenerator)
from aio.run import checker

from envoy.code.check import abstract, typing


YAMLLINT_CONFIG = '.yamllint'
YAMLLINT_MATCH_RE = (
    r"[\w/\.]*\.yml$",
    r"[\w/\.]*\.yaml$", )
YAMLLINT_NOMATCH_RE = (
    r"[\w/\.]*\.template\.yaml$",
    r"[\w/\.]*/server_xds\.cds\.with_unknown_field\.*\.yaml$")


@abstracts.implementer(directory.IDirectoryContext)
class YamllintFilesCheck(directory.ADirectoryContext):

    def __init__(
            self,
            path: str,
            config: YamlLintConfig,
            *args) -> None:
        directory.ADirectoryContext.__init__(self, path)
        self.config = config
        self.args = args

    @property
    def check_results(self) -> Iterator["typing.YamllintProblemTuple"]:
        """Iterate yamllint check results."""
        with self.in_directory:
            for path in self.args:
                if problems := self.run_check(path):
                    yield problems

    def handle_result(
            self,
            path: str,
            result: Generator) -> Optional[
                "typing.YamllintProblemTuple"]:
        if problems := self._parse_problems(path, result):
            return (
                (path,
                 checker.Problems(
                     errors=problems.get("error"),
                     warnings=problems.get("warning"))))

    def run_check(
            self,
            path: str) -> Optional[
                "typing.YamllintProblemTuple"]:
        with io.open(path, newline='') as f:
            return self.handle_result(
                path,
                linter.run(f, self.config, path))

    @debug.logging(
        log=__name__,
        show_cpu=True)
    def run_checks(self) -> Tuple["typing.YamllintProblemTuple", ...]:
        """Run Yamllint checks."""
        return tuple(self.check_results)

    def _parse_problems(
            self,
            path: str,
            problems: Generator) -> Dict[str, List[str]]:
        problem_dict: Dict[str, List[str]] = {}
        for p in problems:
            problem_dict[p.level] = problem_dict.get(p.level, [])
            problem_dict[p.level].append(
                f"{path} ({p.rule} {p.line}:{p.column}): {p.desc}")
        return problem_dict


class AYamllintCheck(abstract.AFileCodeCheck, metaclass=abstracts.Abstraction):

    @classmethod
    def yamllint(
            cls,
            root_path: str,
            config: YamlLintConfig,
            *args) -> Tuple["typing.YamllintProblemTuple", ...]:
        return YamllintFilesCheck(root_path, config, *args).run_checks()

    @async_property
    async def checker_files(self) -> Set[str]:
        """Files with a `.yaml` suffix, but that are not excluded."""
        return set(
            path
            for path
            in await self.directory.files
            if (self.path_match_re.match(path)
                and not self.path_match_exclude_re.match(path)))

    @cached_property
    def config(self) -> YamlLintConfig:
        return YamlLintConfig(file=self.config_path)

    @property
    def config_path(self) -> pathlib.Path:
        return self.directory.path.joinpath(YAMLLINT_CONFIG)

    @cached_property
    def path_match_re(self) -> Pattern[str]:
        """Regex to match files to check."""
        return re.compile("|".join(YAMLLINT_MATCH_RE))

    @cached_property
    def path_match_exclude_re(self) -> Pattern[str]:
        """Regex to match files not to check."""
        return re.compile("|".join(YAMLLINT_NOMATCH_RE))

    @async_property(cache=True)
    async def problem_files(self) -> "typing.ProblemDict":
        return dict(await AwaitableGenerator(self._problem_files))

    @async_property
    async def _problem_files(self) -> AsyncIterator[
            "typing.YamllintProblemTuple"]:
        if not await self.files:
            return

        batches = self.execute_in_batches(
            partial(
                self.yamllint,
                str(self.directory.path),
                self.config),
            *await self.files)

        async for batch in batches:
            for problems in batch:
                yield problems




import asyncio
from concurrent import futures
from typing import Dict, List, Optional, Set

import abstracts

from aio.core import event
from aio.core.directory import ADirectory
from aio.core.functional import async_property

from envoy.base import utils
from envoy.code.check import interface


@abstracts.implementer(event.IExecutive)
class ACodeCheck(event.AExecutive, metaclass=abstracts.Abstraction):

    def __init__(
            self,
            directory: ADirectory,
            fix: bool = False,
            binaries: Optional[Dict[str, str]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        self.directory = directory
        self._fix = fix
        self._loop = loop
        self._pool = pool
        self._binaries = binaries

    @property
    def binaries(self):
        return self._binaries


@abstracts.implementer(interface.IFileCodeCheck)
class AFileCodeCheck(ACodeCheck, metaclass=abstracts.Abstraction):

    @async_property
    @abstracts.interfacemethod
    async def checker_files(self) -> Set[str]:
        raise NotImplementedError

    @async_property(cache=True)
    async def files(self) -> Set[str]:
        files = await self.directory.files
        return (
            files & await self.checker_files
            if files
            else files)

    @property
    def fix(self) -> bool:
        return self._fix

    @async_property
    @abstracts.interfacemethod
    async def problem_files(self) -> Dict[str, List[str]]:
        """Discovered files with flake8 errors."""
        raise NotImplementedError


@abstracts.implementer(interface.IProjectCodeCheck)
class AProjectCodeCheck(ACodeCheck,  metaclass=abstracts.Abstraction):

    def __init__(
            self,
            project: utils.interface.IProject,
            *args,
            **kwargs) -> None:
        self.project = project
        super().__init__(project.directory, *args, **kwargs)




import os
import pathlib
from functools import partial
from typing import AsyncIterator, Iterable, Iterator, Optional, Set, Tuple

import yapf  # type:ignore

import abstracts

from aio.core import directory
from aio.core.dev import debug
from aio.core.functional import (
    async_property,
    AwaitableGenerator)
from aio.run import checker

from envoy.code.check import abstract, interface, typing


YAPF_CONFIG = '.style.yapf'


@abstracts.implementer(directory.IDirectoryContext)
class YapfFormatCheck(directory.ADirectoryContext):
    """Wraps `yapf_api.FormatFile` to run it on multiple paths in a subproc."""

    def __init__(
            self,
            path: str,
            config_path: str,
            fix: bool,
            *args) -> None:
        directory.ADirectoryContext.__init__(self, path)
        self.config_path = config_path
        self.fix = fix
        self.args = args

    @property
    def check_results(self) -> Iterator["typing.YapfProblemTuple"]:
        """Iterate Yapf check results."""
        for path in self.args:
            try:
                result = self.handle_result(
                    path,
                    yapf.yapf_api.FormatFile(
                        os.path.join(self.path, path),
                        style_config=self.config_path,
                        in_place=self.fix,
                        print_diff=not self.fix))
            except yapf.errors.YapfError as e:
                yield path, checker.Problems(
                    errors=[f"Yaml check failed: {path}\n{e}"])
            else:
                if result:
                    yield result

    def handle_result(
            self,
            path: str,
            yapf_result: "typing.YapfResultTuple") -> Optional[
                "typing.YapfProblemTuple"]:
        """Handle a Yapf check result converting to a path and error list."""
        reformatted, encoding, changed = yapf_result
        if not (changed or reformatted):
            return None
        return (
            path,
            (checker.Problems(errors=[f"Issues found (fixed): {path}"])
             if changed and not reformatted
             else checker.Problems(
                     errors=[f"Issues found: {path}\n{reformatted}"])))

    @debug.logging(
        log=__name__,
        show_cpu=True)
    def run_checks(self) -> Tuple["typing.YapfProblemTuple", ...]:
        """Run Yapf checks."""
        return tuple(self.check_results)


@abstracts.implementer(directory.IDirectoryContext)
class YapfFiles(directory.ADirectoryContext):

    @debug.logging(
        log=__name__,
        show_cpu=True)
    def filter_files(self, py_files: Iterable[str]) -> Set[str]:
        with self.in_directory:
            return set(
                yapf.file_resources.GetCommandLineFiles(
                    py_files,
                    recursive=False,
                    exclude=yapf.file_resources.GetExcludePatternsForDir(
                        self.path)))


@abstracts.implementer(interface.IYapfCheck)
class AYapfCheck(abstract.AFileCodeCheck, metaclass=abstracts.Abstraction):

    @classmethod
    def yapf_files(
            cls,
            directory_path: str,
            *py_files: str) -> Set[str]:
        """Yapf file discovery for a given file list."""
        return YapfFiles(directory_path).filter_files(py_files)

    @classmethod
    def yapf_format(
            cls,
            root_path: str,
            config_path: str,
            fix: bool,
            *args) -> "typing.YapfProblemTuple":
        """Run Yapf checks on provided file list."""
        return YapfFormatCheck(root_path, config_path, fix, *args).run_checks()

    @async_property
    async def checker_files(self) -> Set[str]:
        # todo: add grep for py shebang files
        if not (py_files := await self.py_files):
            return set()
        resources = set()
        batches = self.execute_in_batches(
            partial(
                self.yapf_files,
                str(self.directory.path)),
            *py_files)
        async for batch in batches:
            resources |= batch
        return resources

    @property
    def config_path(self) -> pathlib.Path:
        """Path to the Yapf config file."""
        return self.directory.path.joinpath(YAPF_CONFIG)

    @async_property(cache=True)
    async def problem_files(self) -> "typing.ProblemDict":
        return dict(await AwaitableGenerator(self._problem_files))

    @async_property(cache=True)
    async def py_files(self) -> Set[str]:
        """Files with a `.py` suffix."""
        return set(
            path
            for path
            in await self.directory.files
            if path.endswith(".py"))

    @async_property
    async def _problem_files(self) -> AsyncIterator["typing.YapfProblemTuple"]:
        if not await self.files:
            return
        batches = self.execute_in_batches(
            partial(
                self.yapf_format,
                str(self.directory.path),
                str(self.config_path),
                self.fix),
            *await self.files)

        async for batch in batches:
            for path, problem in batch:
                yield path, problem




from .base import ACodeCheck, AFileCodeCheck, AProjectCodeCheck
from .changelog import (
    AChangelogCheck,
    AChangelogChangesChecker,
    AChangelogStatus)
from .checker import ACodeChecker
from .extensions import AExtensionsCheck
from .flake8 import AFlake8Check
from .glint import AGlintCheck
from .rst import (
    ABackticksCheck,
    APunctuationCheck,
    AReflinksCheck)
from .runtime_guards import ARuntimeGuardsCheck
from .shellcheck import AShellcheckCheck
from .yamllint import AYamllintCheck
from .yapf import AYapfCheck
from . import (
    base,
    checker,
    extensions,
    flake8,
    glint,
    shellcheck,
    changelog,
    yamllint,
    yapf)


__all__ = (
    "ABackticksCheck",
    "AChangelogChangesChecker",
    "AChangelogCheck",
    "AChangelogStatus",
    "ACodeCheck",
    "ACodeChecker",
    "AExtensionsCheck",
    "AFileCodeCheck",
    "AFlake8Check",
    "AGlintCheck",
    "AProjectCodeCheck",
    "APunctuationCheck",
    "AReflinksCheck",
    "ARuntimeGuardsCheck",
    "AShellcheckCheck",
    "AYamllintCheck",
    "AYapfCheck",
    "base",
    "checker",
    "flake8",
    "glint",
    "extensions",
    "shellcheck",
    "changelog",
    "yamllint",
    "yapf")




import pathlib
import re
import shutil
import subprocess
from functools import cached_property, partial
from typing import Iterator, List, Optional, Pattern, Set, Tuple, TypedDict

import abstracts

from aio.core import subprocess as _subprocess
from aio.core.functional import async_property
from aio.run import checker

from envoy.code.check import abstract, interface, typing


SHELLCHECK_ERROR_LINE_RE = r"In ([\w\-\_\./]+) line ([0-9]+):"
SHEBANG_RE = (
    r"^#!/bin/bash",
    r"^#!/bin/sh",
    r"^#!/usr/bin/env bash",
    r"^#!/usr/bin/env sh")
SHELLCHECK_MATCH_RE = (
    r"[\w/]*\.sh$", )
SHELLCHECK_NOMATCH_RE = (
    r"[\w/]*\.rst$",
    r"[\w/]*\.md$",
    r"[\w/]*\.genrule_cmd$")


class ShellcheckErrorDict(TypedDict):
    lines: List[str]
    line_numbers: List[int]


@abstracts.implementer(_subprocess.ISubprocessHandler)
class Shellcheck(_subprocess.ASubprocessHandler):

    @cached_property
    def error_line_re(self) -> Pattern[str]:
        """Regex for matching a shellcheck error line.

        This demarcates the beginning of a specific error for a file.
        """
        return re.compile(SHELLCHECK_ERROR_LINE_RE)

    def handle(
            self,
            response: subprocess.CompletedProcess) -> typing.ProblemDict:
        return {}

    def handle_error(
            self,
            response: subprocess.CompletedProcess) -> typing.ProblemDict:
        """Turn the response from a call to shellcheck (for multiple files)
        into an `typing.ProblemDict`."""
        return self._render_errors(self._shellcheck_errors(response))

    def parse_error_line(self, line: str) -> Tuple[str, Optional[int]]:
        """Parse `filename`, `line_number` from a shellcheck error line."""
        matched = self.error_line_re.search(line)
        return (
            (matched.groups()[0],
             int(matched.groups()[1]))
            if matched
            else ("", None))

    def _render_errors(
            self,
            errors: Iterator[
                Tuple[
                    str,
                    ShellcheckErrorDict]]) -> typing.ProblemDict:
        return {
            k: checker.Problems(errors=self._render_file_errors(k, v))
            for k, v
            in errors}

    def _render_file_errors(
            self,
            path: str,
            errors: ShellcheckErrorDict) -> List[str]:
        # This does v basic en pluralization
        line_numbers = ", ".join(str(n) for n in errors["line_numbers"])
        lines = (
            "lines"
            if len(errors["line_numbers"]) > 1
            else "line")
        return [
            "\n".join([
                f"{path} ({lines}: {line_numbers})",
                *errors["lines"]])]

    def _shellcheck_error_info(
            self,
            filename: Optional[str] = None) -> Tuple[
                Optional[str], ShellcheckErrorDict]:
        return (
            filename,
            dict(line_numbers=[], lines=[]))

    def _shellcheck_errors(
            self,
            response: subprocess.CompletedProcess) -> Iterator[
                Tuple[str, ShellcheckErrorDict]]:
        filename, info = self._shellcheck_error_info()
        for line in response.stdout.split("\n"):
            _filename, line_number = self.parse_error_line(line)
            if _filename and line_number:
                # matches as error line - may/not be a new error for this file
                if filename and (_filename != filename):
                    # filename boundary - yield, and reset
                    yield filename, info
                    filename, info = self._shellcheck_error_info(_filename)
                # add line number to error line_numbers
                info["line_numbers"].append(line_number)
                filename = _filename
            if filename is None:
                # preceeding whitespace
                continue
            info["lines"].append(line)
        if filename:
            # flush the buffer
            yield filename, info


@abstracts.implementer(interface.IShellcheckCheck)
class AShellcheckCheck(
        abstract.AFileCodeCheck, metaclass=abstracts.Abstraction):

    @classmethod
    def run_shellcheck(self, path: str, *args) -> typing.ProblemDict:
        """Run shellcheck on files."""
        return Shellcheck(path)(*args)

    @async_property
    async def checker_files(self) -> Set[str]:
        return (
            await self.sh_files
            | await self.shebang_files)

    @cached_property
    def path_match_exclude_re(self) -> Pattern[str]:
        """Regex to match files not to check."""
        return re.compile("|".join(SHELLCHECK_NOMATCH_RE))

    @cached_property
    def path_match_re(self) -> Pattern[str]:
        """Regex to match files to check."""
        return re.compile("|".join(SHELLCHECK_MATCH_RE))

    @async_property(cache=True)
    async def problem_files(self) -> typing.ProblemDict:
        """Discovered shellcheck errors."""
        if not await self.files:
            return {}
        errors: typing.ProblemDict = {}
        jobs = self.execute_in_batches(
            self.shellcheck_executable,
            *await self.files)
        async for result in jobs:
            errors.update(result)
        return errors

    @async_property
    async def sh_files(self) -> Set[str]:
        """Files with a `.sh` suffix, but that are not excluded."""
        return set(
            path
            for path
            in await self.directory.files
            if (self.path_match_re.match(path)
                and not self.path_match_exclude_re.match(path)))

    @async_property
    async def shebang_files(self) -> Set[str]:
        """Files that contain shebang lines, excluding actual .sh files and
        others, eg md/rst, that may have such lines for other reasons."""
        return await self.directory.grep(
            ["-lE", self.shebang_re_expr],
            target=await self._possible_shebang_files)

    @property
    def shebang_re_expr(self) -> str:
        """Regex for matching shebang lines _in_ a file."""
        return "|".join(SHEBANG_RE)

    @cached_property
    def shellcheck_command(self) -> str | pathlib.Path:
        """Shellcheck command, should be available in the running system."""
        if "shellcheck" in self.binaries:
            return pathlib.Path(self.binaries["shellcheck"]).absolute()
        if command := shutil.which("shellcheck"):
            return command
        raise _subprocess.exceptions.OSCommandError(
            "Unable to find shellcheck command")

    @cached_property
    def shellcheck_executable(self) -> partial:
        """Partial with shellcheck command and args."""
        return partial(
            self.run_shellcheck,
            self.directory.path,
            self.shellcheck_command,
            "-x")

    @async_property
    async def _possible_shebang_files(self) -> Set[str]:
        return set(
            path
            for path
            in await self.directory.files
            if (not self.path_match_re.match(path)
                and not self.path_match_exclude_re.match(path)))




import asyncio
import itertools
import json
import logging
import pathlib
import re
from functools import cached_property
from typing import Any, cast, Dict, List, Pattern, Set, Tuple, Type, Union

import yaml as _yaml

import abstracts

from aio.core.functional import async_property

from envoy.base import utils
from envoy.code.check import abstract, exceptions, interface, typing


logger = logging.getLogger(__name__)


FILTER_NAMES_PATTERN = "NetworkFilterNames::get()"
FUZZ_TEST_PATH = (
    "test/extensions/filters/network/common/fuzz/uber_per_readfilter.cc")
METADATA_PATH = "source/extensions/extensions_metadata.yaml"
METADATA_ONLY_EXTENSIONS = (
    "envoy.filters.network.envoy_mobile_http_connection_manager", )
CONTRIB_METADATA_PATH = "contrib/extensions_metadata.yaml"
EXTENSIONS_SCHEMA = "tools/extensions/extensions_schema.yaml"


@abstracts.implementer(interface.IExtensionsCheck)
class AExtensionsCheck(abstract.ACodeCheck, metaclass=abstracts.Abstraction):
    """Extensions check."""

    def __init__(
            self,
            *args,
            **kwargs) -> None:
        self.extensions_build_config = kwargs.pop("extensions_build_config")
        super().__init__(*args, **kwargs)

    @cached_property
    def all_extensions(self) -> Set[str]:
        return (
            set(self.configured_extensions)
            | set(self.builtin_extensions))

    @async_property
    async def all_fuzzed(self) -> bool:
        return (
            await self.robust_to_downstream_count
            == self.fuzzed_count)

    @property
    def builtin_extensions(self) -> typing.ExtensionsSchemaBuiltinList:
        return self.extensions_schema["builtin"]

    @cached_property
    def configured_extensions(self) -> typing.ConfiguredExtensionsDict:
        return cast(
            typing.ConfiguredExtensionsDict,
            self._from_json(
                self.extensions_build_config,
                typing.ConfiguredExtensionsDict,
                "Failed to parse extensions {path}:\n{e}",
                "Extensions parsing error: {path}:\n{e}"))

    @property
    def extension_categories(self) -> typing.ExtensionsSchemaCategoriesList:
        return self.extensions_schema["categories"]

    @property
    def extension_security_postures(
            self) -> typing.ExtensionsSchemaSecurityPosturesList:
        return [
            posture["name"]
            for posture
            in self.extensions_schema["security_postures"]]

    @cached_property
    def extensions_schema(self) -> typing.ExtensionsSchemaDict:
        return cast(
            typing.ExtensionsSchemaDict,
            self._from_yaml(
                self.extensions_schema_path,
                typing.ExtensionsSchemaDict,
                "Failed to parse extensions schema {path}:\n{e}",
                "Extensions schema parsing error: {path}:\n{e}"))

    @property
    def extensions_schema_path(self) -> pathlib.Path:
        return self.directory.path.joinpath(EXTENSIONS_SCHEMA)

    @property
    def extension_status_values(
            self) -> typing.ExtensionsSchemaStatusValuesList:
        return [
            posture["name"]
            for posture
            in self.extensions_schema["status_values"]]

    @property
    def fuzz_test_path(self) -> pathlib.Path:
        return self.directory.path.joinpath(FUZZ_TEST_PATH)

    @property
    def fuzzed_count(self) -> int:
        # Hack-ish! We only search the first 50 lines to capture the filters
        # in `filterNames()`.
        return len(
            self.fuzzed_filter_names_re.findall(
                "".join(self.fuzz_test_path.read_text().splitlines()[:50])))

    @cached_property
    def fuzzed_filter_names_re(self) -> Pattern:
        return re.compile(FILTER_NAMES_PATTERN)

    @async_property(cache=True)
    async def metadata(self) -> typing.ExtensionsMetadataDict:
        return dict(**await self.metadata_core, **await self.metadata_contrib)

    @async_property
    async def metadata_contrib(self) -> typing.ExtensionsMetadataDict:
        return await self._metadata(self.metadata_contrib_path)

    @property
    def metadata_contrib_path(self) -> pathlib.Path:
        return self.directory.path.joinpath(CONTRIB_METADATA_PATH)

    @async_property
    async def metadata_core(self) -> typing.ExtensionsMetadataDict:
        return await self._metadata(self.metadata_core_path)

    @property
    def metadata_core_path(self) -> pathlib.Path:
        return self.directory.path.joinpath(METADATA_PATH)

    @async_property
    async def metadata_errors(self) -> Dict[str, Tuple[str, ...]]:
        return {
            extension: await self.check_metadata(extension)
            for extension
            in await self.metadata}

    @async_property
    async def metadata_missing(self) -> Set[str]:
        return (
            self.all_extensions
            - set(await self.metadata))

    @async_property
    async def metadata_only(self) -> Set[str]:
        return (
            set(await self.metadata)
            - self.metadata_only_extensions
            - self.all_extensions)

    @property
    def metadata_only_extensions(self) -> Set[str]:
        return set(METADATA_ONLY_EXTENSIONS)

    @async_property
    async def registration_errors(self) -> List[str]:
        return [
            *[f"Metadata for unused extension found: {extension}"
              for extension
              in sorted(await self.metadata_only)],
            *[f"Metadata missing for extension: {extension}"
              for extension
              in sorted(await self.metadata_missing)]]

    @async_property
    async def robust_to_downstream_count(self) -> int:
        # Count number of network filters robust to untrusted downstreams.
        return len([
            ext
            for ext, data
            in (await self.metadata).items()
            if ("network" in ext
                and (data["security_posture"]
                     == "robust_to_untrusted_downstream"))])

    async def check_metadata(self, extension: str) -> Tuple[str, ...]:
        return tuple(
            itertools.chain.from_iterable(
                await asyncio.gather(
                    self._check_metadata_categories(extension),
                    self._check_metadata_security_posture(extension),
                    self._check_metadata_status(extension))))

    async def _check_metadata_categories(
            self, extension: str) -> Tuple[str, ...]:
        categories = (await self.metadata)[extension].get("categories", ())
        if not categories:
            return (
                f"Missing extension category for {extension}. "
                "Please make sure the target is an envoy_cc_extension "
                "and category is set", )
        return tuple(
            (f"Unknown extension category for {extension}: {cat}. "
             "Please add it to tools/extensions/extensions_check.py")
            for cat
            in categories
            if cat not in self.extension_categories)

    async def _check_metadata_security_posture(
            self, extension: str) -> Tuple[str, ...]:
        security_posture = (await self.metadata)[extension]["security_posture"]
        if not security_posture:
            return (
                f"Missing security posture for {extension}. "
                "Please make sure the target is an "
                "envoy_cc_extension and security_posture is set", )
        elif security_posture not in self.extension_security_postures:
            return (
                "Unknown security posture for "
                f"{extension}: {security_posture}", )
        return ()

    async def _check_metadata_status(
            self, extension: str) -> Tuple[str, ...]:
        status = (await self.metadata)[extension]["status"]
        if status not in self.extension_status_values:
            return (f"Unknown status for {extension}: {status}", )
        return ()

    def _from_json(
            self,
            path: Union[str, pathlib.Path],
            type: Type,
            err_message: str,
            warn_message: str) -> Any:
        # Parse JSON, handling errors
        try:
            return utils.from_json(path, type)
        except (json.JSONDecodeError, FileNotFoundError) as e:
            raise exceptions.ExtensionsConfigurationError(
                err_message.format(path=path, e=e))
        except utils.TypeCastingError as e:
            logger.warning(warn_message.format(path=path, e=e))
            return e.value

    def _from_yaml(
            self,
            path: Union[str, pathlib.Path],
            type: Type,
            err_message: str,
            warn_message: str) -> Any:
        # Parse YAML, handling errors
        try:
            return utils.from_yaml(path, type)
        except (_yaml.reader.ReaderError, FileNotFoundError) as e:
            raise exceptions.ExtensionsConfigurationError(
                err_message.format(path=path, e=e))
        except utils.TypeCastingError as e:
            logger.warning(warn_message.format(path=path, e=e))
            return e.value

    async def _metadata(self, path) -> typing.ExtensionsMetadataDict:
        try:
            return await self.execute(
                utils.from_yaml,
                path,
                typing.ExtensionsMetadataDict)
        except (utils.exceptions.TypeCastingError, FileNotFoundError) as e:
            raise exceptions.ExtensionsConfigurationError(
                "Failed to parse extensions metadata "
                f"({path}): {e}")



"""Abstract code checker."""

import abc
import argparse
import pathlib
import re
from functools import cached_property
from typing import Dict, Mapping, Optional, Pattern, Set, Type

from yamllint.config import YamlLintConfigError  # type:ignore

import abstracts

from aio.core import directory as _directory, event, subprocess
from aio.core.tasks import inflate
from aio.run import checker

from envoy.base import utils
from envoy.base.utils import IProject
from envoy.code.check import exceptions, interface, typing


# TODO: Add a README in envoy repo with info on how to fix and maybe use
#   a template
GLINT_ADVICE = (
    "Glint check failed\n"
    "\n"
    "  Please fix your editor to ensure:\n"
    "\n"
    "      - no trailing whitespace\n"
    "      - no preceding mixed tabs/spaces\n"
    "      - all files end with a newline")

NO_EXTENSIONS_ERROR_MSG = (
    "`--extensions_build_config` not provided, disabling extensions checks")


class CodeCheckerSummary(checker.CheckerSummary):

    def print_summary(self) -> None:
        """Write summary to stderr."""
        super().print_summary()
        if "glint" in self.checker.errors:
            self.writer_for("error")(GLINT_ADVICE)


@abstracts.implementer(interface.ICodeChecker)
class ACodeChecker(
        checker.Checker,
        event.AReactive,
        metaclass=abstracts.Abstraction):
    """Code checker."""

    checks = (
        "changelog",
        "extensions_fuzzed",
        "extensions_metadata",
        "extensions_registered",
        "glint",
        "python_yapf",
        "python_flake8",
        "runtime_guards",
        "shellcheck",
        "yamllint")

    @property
    def all_files(self) -> bool:
        return self.args.all_files

    @cached_property
    def binaries(self) -> Dict[str, str]:
        return dict(
            binary.split(":")
            for binary
            in self.args.binary or [])

    @cached_property
    def disabled_checks(self):
        disabled = {}
        if not self.args.extensions_build_config:
            disabled["extensions_fuzzed"] = NO_EXTENSIONS_ERROR_MSG
            disabled["extensions_metadata"] = NO_EXTENSIONS_ERROR_MSG
            disabled["extensions_registered"] = NO_EXTENSIONS_ERROR_MSG
        return disabled

    @property
    def changed_since(self) -> Optional[str]:
        return self.args.since

    @cached_property
    def changelog(self) -> "interface.IChangelogCheck":
        """Changelog checker."""
        return self.changelog_class(
            self.project,
            **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def changelog_class(self) -> Type["interface.IChangelogCheck"]:
        raise NotImplementedError

    @cached_property
    def check_kwargs(self) -> Mapping:
        return dict(
            fix=self.fix,
            binaries=self.binaries,
            loop=self.loop,
            pool=self.pool)

    @cached_property
    def directory(self) -> "_directory.ADirectory":
        """Greppable directory - optionally in a git repo, depending on whether
        we want to look at all files.
        """
        return self.project.directory.filtered(**self.directory_kwargs)

    @property
    def directory_kwargs(self) -> Dict:
        kwargs: Dict = dict(
            exclude_matcher=self.grep_excluding_re,
            path_matcher=self.grep_matching_re,
            untracked=self.all_files)
        if not self.all_files:
            kwargs["changed"] = self.changed_since
        return kwargs

    @cached_property
    def extensions(self) -> "interface.IExtensionsCheck":
        """Extensions checker."""
        return self.extensions_class(
            self.directory,
            extensions_build_config=self.args.extensions_build_config,
            **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def extensions_class(self) -> Type["interface.IExtensionsCheck"]:
        raise NotImplementedError

    @cached_property
    def flake8(self) -> "interface.IFlake8Check":
        """Flake8 checker."""
        return self.flake8_class(self.directory, **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def flake8_class(self) -> Type["interface.IFlake8Check"]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def fs_directory_class(self) -> Type["_directory.ADirectory"]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def git_directory_class(self) -> Type["_directory.AGitDirectory"]:
        raise NotImplementedError

    @cached_property
    def glint(self) -> "interface.IGlintCheck":
        """Glint checker."""
        return self.glint_class(self.directory, **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def glint_class(self) -> Type["interface.IGlintCheck"]:
        raise NotImplementedError

    @property
    def grep_excluding_re(self) -> Optional[Pattern[str]]:
        return self._grep_re(self.args.excluding)

    @property
    def grep_matching_re(self) -> Optional[Pattern[str]]:
        return self._grep_re(self.args.matching)

    @property
    @abc.abstractmethod
    def path(self) -> pathlib.Path:
        return super().path

    @cached_property
    def project(self) -> IProject:
        return self.project_class(self.path)

    @property  # type:ignore
    @abstracts.interfacemethod
    def project_class(self) -> Type[IProject]:
        raise NotImplementedError

    @cached_property
    def runtime_guards(self) -> "interface.IRuntimeGuardsCheck":
        """Shellcheck checker."""
        return self.runtime_guards_class(
            self.project,
            **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def runtime_guards_class(self) -> Type["interface.IRuntimeGuardsCheck"]:
        raise NotImplementedError

    @cached_property
    def shellcheck(self) -> "interface.IShellcheckCheck":
        """Shellcheck checker."""
        return self.shellcheck_class(self.directory, **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def shellcheck_class(self) -> Type["interface.IShellcheckCheck"]:
        raise NotImplementedError

    @property
    def summary_class(self) -> Type[CodeCheckerSummary]:
        """CodeChecker's summary class."""
        return CodeCheckerSummary

    @cached_property
    def yamllint(self) -> interface.IYamllintCheck:
        """Yamllint checker."""
        return self.yamllint_class(self.directory, **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def yamllint_class(self) -> Type[interface.IYamllintCheck]:
        raise NotImplementedError

    @cached_property
    def yapf(self) -> interface.IYapfCheck:
        """YAPF checker."""
        return self.yapf_class(self.directory, **self.check_kwargs)

    @property  # type:ignore
    @abstracts.interfacemethod
    def yapf_class(self) -> Type[interface.IYapfCheck]:
        raise NotImplementedError

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("-a", "--all_files", action="store_true")
        parser.add_argument("-m", "--matching", action="append")
        parser.add_argument("-x", "--excluding", action="append")
        parser.add_argument("-b", "--binary", action="append")
        parser.add_argument("-s", "--since")
        parser.add_argument("--extensions_build_config")

    async def check_changelog(self):
        for changelog in self.changelog:
            if errors := await changelog.errors:
                self.error("changelog", errors)
            else:
                self.succeed(
                    "changelog",
                    [f"{changelog.version}"])

    async def check_extensions_fuzzed(self) -> None:
        """Check for glint issues."""
        if await self.extensions.all_fuzzed:
            self.succeed(
                "extensions_fuzzed",
                ["All network filters are fuzzed"])
        else:
            self.error(
                "extensions_fuzzed",
                ["Check that all network filters robust against untrusted "
                 "downstreams are fuzzed by adding them to filterNames() "
                 f"in {self.extensions.fuzz_test_path}"])

    async def check_extensions_metadata(self) -> None:
        """Check for glint issues."""
        errors = await self.extensions.metadata_errors
        for extension, errors in errors.items():
            if errors:
                self.error("extensions_metadata", errors)
            else:
                self.succeed("extensions_metadata", [f"{extension}"])

    async def check_extensions_registered(self) -> None:
        """Check for glint issues."""
        if errors := await self.extensions.registration_errors:
            self.error("extensions_registered", errors)
        else:
            self.succeed(
                "extensions_registered",
                ["Registered metadata matches found extensions"])

    async def check_glint(self) -> None:
        """Check for glint issues."""
        await self._code_check(self.glint)

    # TODO: catch errors in checkers as well as preloaders
    async def check_python_flake8(self) -> None:
        """Check for flake8 issues."""
        await self._code_check(self.flake8)

    async def check_python_yapf(self) -> None:
        """Check for yapf issues."""
        await self._code_check(self.yapf)

    async def check_runtime_guards(self) -> None:
        """Check for shellcheck issues."""
        async for guard, status in self.runtime_guards.status:
            if status is None:
                self.log.info(f"Ignoring runtime guard: {guard}")
            elif not status:
                self.error(
                    "runtime_guards",
                    [f"Missing from changelogs: {guard}"])
            else:
                self.succeed(
                    "runtime_guards",
                    [f"In changelogs: {guard}"])

    async def check_shellcheck(self) -> None:
        """Check for shellcheck issues."""
        await self._code_check(self.shellcheck)

    async def check_yamllint(self) -> None:
        """Check for yamllint issues."""
        await self._code_check(self.yamllint)

    @checker.preload(
        when=["changelog"],
        catches=[utils.exceptions.ChangelogParseError])
    async def preload_changelog(self) -> None:
        preloader = inflate(
            self.changelog,
            lambda c: (c.errors, ))
        async for changelog in preloader:
            self.log.debug(f"Preloaded changelog: {changelog.version}")

    @checker.preload(
        when=["extensions_fuzzed",
              "extensions_metadata",
              "extensions_registered"],
        catches=[exceptions.ExtensionsConfigurationError])
    async def preload_extensions(self) -> None:
        metadata = await self.extensions.metadata
        self.extensions.extensions_schema
        self.log.debug(f"Preloaded extensions ({len(metadata)})")

    @checker.preload(
        when=["python_flake8"],
        catches=[subprocess.exceptions.OSCommandError])
    async def preload_flake8(self) -> None:
        await self.flake8.problem_files

    @checker.preload(
        when=["glint"],
        catches=[subprocess.exceptions.OSCommandError])
    async def preload_glint(self) -> None:
        await self.glint.problem_files

    @checker.preload(
        when=["shellcheck"],
        catches=[subprocess.exceptions.OSCommandError])
    async def preload_shellcheck(self) -> None:
        await self.shellcheck.problem_files

    @checker.preload(
        when=["runtime_guards"],
        catches=[subprocess.exceptions.OSCommandError])
    async def preload_runtime_guards(self) -> None:
        await self.runtime_guards.missing

    @checker.preload(
        when=["yamllint"],
        catches=[
            subprocess.exceptions.OSCommandError,
            YamlLintConfigError])
    async def preload_yamllint(self) -> None:
        await self.yamllint.problem_files

    @checker.preload(
        when=["python_yapf"],
        catches=[subprocess.exceptions.OSCommandError])
    async def preload_yapf(self) -> None:
        await self.yapf.problem_files

    def _check_output(
            self,
            check_files: Set[str],
            problem_files: typing.ProblemDict) -> None:
        # This can be slow/blocking for large result sets, run
        # in a separate thread
        for path in sorted(check_files):
            if path not in problem_files:
                self.succeed(
                    self.active_check,
                    [path])
                continue
            if problem_files[path].errors:
                self.error(
                    self.active_check,
                    problem_files[path].errors)
            if problem_files[path].warnings:
                self.warn(
                    self.active_check,
                    problem_files[path].warnings)

    async def _code_check(self, check: "interface.IFileCodeCheck") -> None:
        await self.loop.run_in_executor(
            None,
            self._check_output,
            await check.files,
            await check.problem_files)

    def _grep_re(self, arg: Optional[str]) -> Optional[Pattern[str]]:
        # When using system `grep` we want to filter out at least some
        # of the files that .gitignore would.
        # TODO: use globs on cli and covert to re here
        return (
            re.compile("|".join(arg))
            if arg
            else None)




from typing import Dict, List, Optional, Tuple, TypedDict

from aio.run import checker


ProblemDict = Dict[str, checker.interface.IProblems]

YamllintProblemTuple = Tuple[str, checker.interface.IProblems]

YapfProblemTuple = Tuple[str, checker.interface.IProblems]
YapfResultTuple = Tuple[str, str, bool]
YapfCheckResultTuple = Tuple[str, YapfResultTuple]


class BaseExtensionMetadataDict(TypedDict):
    categories: List[str]
    security_posture: str
    status: str


class ExtensionMetadataDict(BaseExtensionMetadataDict, total=False):
    undocumented: bool
    type_urls: List[str]


ExtensionsMetadataDict = Dict[str, ExtensionMetadataDict]
ExtensionsSchemaCategoriesList = List[str]
ExtensionsSchemaBuiltinList = List[str]
ExtensionsSchemaSecurityPosturesList = List[str]
ExtensionsSchemaStatusValuesList = List[str]


class ExtensionSecurityPostureMetadataDict(TypedDict):
    name: str
    description: str


class ExtensionStatusMetadataDict(TypedDict):
    name: str
    description: Optional[str]


class ExtensionsSchemaDict(TypedDict):
    builtin: ExtensionsSchemaBuiltinList
    security_postures: List[ExtensionSecurityPostureMetadataDict]
    categories: ExtensionsSchemaCategoriesList
    status_values: List[ExtensionStatusMetadataDict]


ConfiguredExtensionsDict = Dict[str, str]




import pathlib
from functools import cached_property
from typing import Tuple, Type

import abstracts

from aio.core import directory

from envoy.base.utils import IProject, Project
from envoy.code.check import abstract, interface


@abstracts.implementer(interface.IExtensionsCheck)
class ExtensionsCheck(abstract.AExtensionsCheck):
    pass


@abstracts.implementer(interface.IFlake8Check)
class Flake8Check(abstract.AFlake8Check):
    pass


@abstracts.implementer(interface.IGlintCheck)
class GlintCheck(abstract.AGlintCheck):
    pass


@abstracts.implementer(interface.IShellcheckCheck)
class ShellcheckCheck(abstract.AShellcheckCheck):
    pass


@abstracts.implementer(interface.IYapfCheck)
class YapfCheck(abstract.AYapfCheck):
    pass


@abstracts.implementer(abstract.AYamllintCheck)
class YamllintCheck:
    pass


@abstracts.implementer(interface.IRSTCheck)
class BackticksCheck(abstract.ABackticksCheck):
    pass


@abstracts.implementer(interface.IRSTCheck)
class PunctuationCheck(abstract.APunctuationCheck):
    pass


@abstracts.implementer(interface.IRSTCheck)
class ReflinksCheck(abstract.AReflinksCheck):
    pass


@abstracts.implementer(interface.IRuntimeGuardsCheck)
class RuntimeGuardsCheck(abstract.ARuntimeGuardsCheck):
    pass


@abstracts.implementer(interface.IChangelogChangesChecker)
class ChangelogChangesChecker(abstract.AChangelogChangesChecker):

    @cached_property
    def change_checkers(self) -> Tuple[interface.IRSTCheck, ...]:
        return (
            BackticksCheck(),
            PunctuationCheck(),
            ReflinksCheck())


@abstracts.implementer(interface.IChangelogStatus)
class ChangelogStatus(abstract.AChangelogStatus):
    pass


@abstracts.implementer(interface.IChangelogCheck)
class ChangelogCheck(abstract.AChangelogCheck):

    @property
    def changes_checker_class(
            self) -> Type[interface.IChangelogChangesChecker]:
        return ChangelogChangesChecker

    @property
    def changelog_status_class(self) -> Type[interface.IChangelogStatus]:
        return ChangelogStatus


@abstracts.implementer(interface.ICodeChecker)
class CodeChecker(abstract.ACodeChecker):

    @property
    def extensions_class(self):
        return ExtensionsCheck

    @property
    def flake8_class(self):
        return Flake8Check

    @property
    def fs_directory_class(self):
        return directory.Directory

    @property
    def git_directory_class(self):
        return directory.GitDirectory

    @property
    def glint_class(self):
        return GlintCheck

    @cached_property
    def path(self) -> pathlib.Path:
        return super().path

    @property
    def project_class(self) -> Type[IProject]:
        return Project

    @property
    def runtime_guards_class(self):
        return RuntimeGuardsCheck

    @property
    def shellcheck_class(self):
        return ShellcheckCheck

    @property
    def changelog_class(self):
        return ChangelogCheck

    @property
    def yapf_class(self):
        return YapfCheck

    @property
    def yamllint_class(self):
        return YamllintCheck




import re
from functools import cached_property
from typing import List, Optional, Pattern

import abstracts

from envoy.code.check import interface


INVALID_REFLINK_RE = r"[^:]ref:`|^ref:`"
PUNCTUATION_RE = r".*\. <[^<]+>`$|.*\.$"

# Make sure backticks come in pairs.
# Exceptions: reflinks (ref:`` where the backtick won't be preceded by a space
#             links `title <link>`_ where the _ is checked for in the regex.
SINGLE_TICK_RE = r"[^`:]`[^`]+`[^`]"
REF_TICKS_RE = r":`[^`]+`"
LINK_TICKS_RE = r"[^`]`[^`]+`_"


@abstracts.implementer(interface.IRSTCheck)
class ABackticksCheck(metaclass=abstracts.Abstraction):
    error_message = "Single backticks found {single_ticks}"

    def __call__(self, text: str) -> Optional[str]:
        single_ticks = self._find_single_ticks(text)
        return (
            self.error_message.format(single_ticks=', '.join(single_ticks))
            if single_ticks
            else None)

    @cached_property
    def link_ticks_re(self) -> Pattern[str]:
        return re.compile(LINK_TICKS_RE)

    @cached_property
    def ref_ticks_re(self) -> Pattern[str]:
        return re.compile(REF_TICKS_RE)

    @cached_property
    def single_tick_re(self) -> Pattern[str]:
        return re.compile(SINGLE_TICK_RE)

    def _find_single_ticks(
            self,
            text: str) -> List[str]:
        return [
            bad_ticks[1:-1]
            for bad_ticks
            in self.single_tick_re.findall(
                self._strip_valid_refs(text))]

    def _strip_valid_refs(self, text: str) -> str:
        for reflink in self.ref_ticks_re.findall(text):
            text = text.replace(reflink, "")
        for extlink in self.link_ticks_re.findall(text):
            text = text.replace(extlink, "")
        return text


@abstracts.implementer(interface.IRSTCheck)
class AReflinksCheck(metaclass=abstracts.Abstraction):
    error_message = "Invalid ref link `ref:` should be `:ref:`"

    def __call__(self, text: str) -> Optional[str]:
        return (
            self.error_message
            if self.invalid_reflink_re.findall(text)
            else None)

    @cached_property
    def invalid_reflink_re(self) -> Pattern[str]:
        return re.compile(INVALID_REFLINK_RE)


@abstracts.implementer(interface.IRSTCheck)
class APunctuationCheck(metaclass=abstracts.Abstraction):
    error_message = "Missing punctuation \"...{snippet}\""

    def __call__(self, text: str) -> Optional[str]:
        return (
            self.error_message.format(snippet=text[-30:])
            if not self._check_punctuation(text)
            else None)

    @cached_property
    def punctuation_re(self) -> Pattern[str]:
        return re.compile(PUNCTUATION_RE, re.DOTALL)

    def _check_punctuation(
            self,
            text: str) -> bool:
        return bool(
            self.punctuation_re.match(text)
            # Ends with a list
            or text.split("\n")[-1].startswith("  *"))




import itertools
from datetime import datetime
from functools import cached_property
from typing import Iterator, Optional, Tuple, Type

from packaging import version as _version

import abstracts

from aio.core.functional import async_property

from envoy.base import utils
from envoy.code.check import abstract, interface


MAX_VERSION_FOR_CHANGES_SECTION = "1.16"


@abstracts.implementer(interface.IChangelogChangesChecker)
class AChangelogChangesChecker(metaclass=abstracts.Abstraction):
    error_message = (
        "{version}/{section}/{entry[area]}: "
        "{error}\n{entry[change]}")

    def __init__(
            self,
            sections: utils.typing.ChangelogSectionsDict) -> None:
        self.sections = sections

    @property  # type:ignore
    @abstracts.interfacemethod
    def change_checkers(self) -> Tuple[interface.IRSTCheck, ...]:
        raise NotImplementedError

    @cached_property
    def max_version_for_changes_section(self) -> _version.Version:
        return _version.Version(MAX_VERSION_FOR_CHANGES_SECTION)

    def check_entry(
            self,
            version: _version.Version,
            section: str,
            entry: utils.typing.ChangeDict) -> Tuple[str, ...]:
        change = entry["change"].strip()
        errors = [
            checker(change)
            for checker
            in self.change_checkers]
        return tuple(
            self.error_message.format(
                version=version,
                section=section,
                entry=entry,
                error=error)
            for error
            in errors
            if error)

    def check_section(
            self,
            version: _version.Version,
            section: str,
            data: Optional[utils.typing.ChangeList]) -> Tuple[str, ...]:
        name_error = self.check_section_name(version, section)
        return (
            *((name_error, )
              if name_error
              else ()),
            *itertools.chain.from_iterable(
                self.check_entry(version, section, entry)
                for entry
                in data or []))

    def check_sections(
            self,
            version: _version.Version,
            sections: utils.typing.ChangelogChangeSectionsDict) -> (
                Tuple[str, ...]):
        return tuple(
            itertools.chain.from_iterable(
                self.check_section(version, section, data)  # type:ignore
                for section, data
                in sections.items()))

    def check_section_name(
            self,
            version: _version.Version,
            section: str) -> Optional[str]:
        invalid_changes = (
            section == "changes"
            and version > self.max_version_for_changes_section)
        if invalid_changes:
            return (
                f"{version}/changes: Invalid `changes` section "
                "(this is no longer used)")


@abstracts.implementer(interface.IChangelogStatus)
class AChangelogStatus(metaclass=abstracts.Abstraction):

    def __init__(
            self,
            check: interface.IChangelogCheck,
            changelog: utils.interface.IChangelog) -> None:
        self._check = check
        self.changelog = changelog

    @property
    def checker(self) -> interface.IChangelogChangesChecker:
        return self._check.changes_checker

    @async_property
    async def data(self) -> utils.typing.ChangelogDict:
        return await self.changelog.data

    @async_property
    async def date(self) -> str:
        return (await self.data)["date"]

    @cached_property
    def date_format(self) -> str:
        return self.project.changelogs.date_format.replace("-", "")

    @async_property
    async def dev_not_pending(self) -> bool:
        return (
            self.is_current
            and self.project.is_dev
            and not await self.is_pending)

    @property
    def duplicate_current(self) -> bool:
        return (
            self.is_current
            and self.project.changelogs.changelog_path(
                self.version).exists())

    @async_property(cache=True)
    async def errors(self) -> Tuple[str, ...]:
        try:
            return (
                *self.check_version(),
                *await self.check_date(),
                *await self.check_sections())
        except utils.exceptions.ChangelogParseError as e:
            return (f"{self.version}: {e}", )

    @async_property
    async def invalid_date(self) -> Optional[str]:
        if await self.is_pending:
            return None
        date = await self.date
        try:
            datetime.strptime(date, self.date_format)
        except ValueError:
            return date

    @cached_property
    def is_current(self) -> bool:
        return self.project.is_current(self.version)

    @async_property
    async def is_pending(self) -> bool:
        return (await self.date) == "Pending"

    @async_property
    async def pending_not_dev(self) -> bool:
        return (
            (not self.is_current
             or not self.project.is_dev)
            and await self.is_pending)

    @property
    def project(self) -> utils.interface.IProject:
        return self._check.project

    @async_property
    async def sections(self) -> utils.typing.ChangelogChangeSectionsDict:
        return utils.typed(
            utils.typing.ChangelogChangeSectionsDict,
            {k: v
             for k, v
             in (await self.data).items()
             if k != "date"})

    @property
    def version(self) -> _version.Version:
        return self.changelog.version

    @property
    def version_higher_than_current(self) -> bool:
        return (
            self.version
            > _version.Version(self.project.version.base_version))

    async def check_date(self) -> Tuple[str, ...]:
        errors = []
        if invalid_date := await self.invalid_date:
            errors.append(f"Format not recognized \"{invalid_date}\"")
        if await self.dev_not_pending:
            errors.append("Should be set to `Pending`")
        elif await self.pending_not_dev:
            errors.append("Should not be set to `Pending`")
        return tuple(
            f"{self.version}/date: {e}"
            for e
            in errors)

    async def check_sections(self) -> Tuple[str, ...]:
        # Runs checker in executor, uncomment following line for debugging
        # return self.checker.check_sections(self.version, await self.sections)
        return await self.project.execute(
            self.checker.check_sections,
            self.version,
            await self.sections)

    def check_version(self) -> Tuple[str, ...]:
        errors = []
        if self.duplicate_current:
            errors.append(
                "Duplicate current version file. "
                "Only `current.yaml` should exist for the current version "
                f"({self.project.version.base_version})")
        elif self.version_higher_than_current:
            errors.append(
                "Changelog version is higher than "
                f"the current version ({self.project.version.base_version})")
        return tuple(
            f"{self.version}: {e}"
            for e
            in errors)


@abstracts.implementer(interface.IChangelogCheck)
class AChangelogCheck(
        abstract.AProjectCodeCheck,
        metaclass=abstracts.Abstraction):
    """Changelog check."""

    def __iter__(self) -> Iterator[interface.IChangelogStatus]:
        for changelog in self.changelogs:
            yield changelog

    @property  # type:ignore
    @abstracts.interfacemethod
    def changes_checker_class(
            self) -> Type[interface.IChangelogChangesChecker]:
        raise NotImplementedError

    @cached_property
    def changes_checker(self) -> interface.IChangelogChangesChecker:
        return self.changes_checker_class(
            self.project.changelogs.sections)

    @property  # type:ignore
    @abstracts.interfacemethod
    def changelog_status_class(self) -> Type[interface.IChangelogStatus]:
        raise NotImplementedError

    @cached_property
    def changelogs(self) -> Tuple[interface.IChangelogStatus, ...]:
        return tuple(
            self.changelog_status_class(self, changelog)
            for changelog
            in self.project.changelogs.values())




from . import abstract, exceptions, typing
from .abstract import (
    ABackticksCheck,
    AChangelogChangesChecker,
    AChangelogCheck,
    AChangelogStatus,
    ACodeCheck,
    ACodeChecker,
    AExtensionsCheck,
    AFlake8Check,
    AFileCodeCheck,
    AGlintCheck,
    AProjectCodeCheck,
    APunctuationCheck,
    AReflinksCheck,
    ARuntimeGuardsCheck,
    AShellcheckCheck,
    AYamllintCheck,
    AYapfCheck)
from .checker import (
    ChangelogChangesChecker,
    ChangelogCheck,
    ChangelogStatus,
    CodeChecker,
    ExtensionsCheck,
    Flake8Check,
    GlintCheck,
    RuntimeGuardsCheck,
    ShellcheckCheck,
    YamllintCheck,
    YapfCheck)
from .cmd import run, main
from . import checker, interface


__all__ = (
    "abstract",
    "ABackticksCheck",
    "AChangelogChangesChecker",
    "AChangelogCheck",
    "AChangelogStatus",
    "ACodeCheck",
    "ACodeChecker",
    "AExtensionsCheck",
    "AFileCodeCheck",
    "AFlake8Check",
    "AGlintCheck",
    "AProjectCodeCheck",
    "APunctuationCheck",
    "AReflinksCheck",
    "ARuntimeGuardsCheck",
    "AShellcheckCheck",
    "AYamllintCheck",
    "AYapfCheck",
    "ChangelogChangesChecker",
    "ChangelogCheck",
    "ChangelogStatus",
    "checker",
    "CodeChecker",
    "exceptions",
    "ExtensionsCheck",
    "Flake8Check",
    "GlintCheck",
    "interface",
    "main",
    "run",
    "main",
    "run",
    "RuntimeGuardsCheck",
    "ShellcheckCheck",
    "typing",
    "YamllintCheck",
    "YapfCheck")




import sys

from .checker import CodeChecker


def main(*args: str) -> int:
    return CodeChecker(*args)()


def run() -> None:
    sys.exit(main(*sys.argv[1:]))


if __name__ == "__main__":
    run()




import re
from functools import cached_property
from typing import (
    AsyncIterator, Awaitable, Iterable, Optional, Pattern, Set, Tuple)

import abstracts

from aio.core.functional import async_property

from envoy.code.check import abstract, interface


EXPECTED_MISSING_GUARDS = (
    "envoy_reloadable_features_admin_stats_filter_use_re2",
    "envoy_reloadable_features_allow_multiple_dns_addresses",
    "envoy_reloadable_features_get_route_config_factory_by_type",
    "envoy_reloadable_features_internal_address",
    "envoy_reloadable_features_local_ratelimit_match_all_descriptors",
    "envoy_reloadable_features_postpone_h3_client_connect_to_next_loop",
    "envoy_reloadable_features_test_feature_true")
RELOADABLE_GUARD_GREP_RE = r"^RUNTIME_GUARD\(envoy_reloadable"
RELOADABLE_MATCH_RE = r"``envoy.reloadable[._][a-z0-9_.]+``"
RUNTIME_GUARDS_CONFIG_PATH = "source/common/runtime/runtime_features.cc"


@abstracts.implementer(interface.IRuntimeGuardsCheck)
class ARuntimeGuardsCheck(
        abstract.AProjectCodeCheck,
        metaclass=abstracts.Abstraction):
    """Runtime guards check."""

    @async_property(cache=True)
    async def configured(self) -> Set[str]:
        return set(
            line.split(":")[1][14:-2]
            for line
            in await self._grepped)

    @cached_property
    def expected_missing(self) -> Set[str]:
        return set(EXPECTED_MISSING_GUARDS)

    @async_property
    async def mentioned(self) -> Set[str]:
        mentioned = set()
        async for change in self._changes:
            mentioned |= self._find_mention(change)
        return mentioned

    @async_property(cache=True)
    async def missing(self) -> Set[str]:
        return set(
            await self.configured
            - await self.mentioned
            - self.expected_missing)

    @cached_property
    def reloadable_match_re(self) -> Pattern:
        return re.compile(RELOADABLE_MATCH_RE)

    @async_property
    async def status(self) -> AsyncIterator[Tuple[str, Optional[bool]]]:
        for guard in sorted(await self.configured):
            if guard in await self.missing:
                yield guard, False
            elif guard in self.expected_missing:
                yield guard, None
            else:
                yield guard, True

    @async_property
    async def _changes(self) -> AsyncIterator[str]:
        for changelog in self.project.changelogs.values():
            for section, data in (await changelog.data).items():
                if section == "date":
                    continue
                for change in data:
                    yield change["change"]

    @property
    def _grepped(self) -> Awaitable[Iterable[str]]:
        return self.directory.grep(
            ["-E", RELOADABLE_GUARD_GREP_RE],
            RUNTIME_GUARDS_CONFIG_PATH)

    def _find_mention(self, change: str) -> Set[str]:
        return set(
            m.strip("`").replace(".", "_")
            for m
            in self.reloadable_match_re.findall(change))




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: envoy.code.check:package

from setuptools import setup

setup(**{
    'author': 'Ryan Northey',
    'author_email': 'ryan@synca.io',
    'description': '"Code checker used in Envoy proxy\'s CI"',
    'entry_points': {
        'console_scripts': [
            'envoy.code.check=envoy.code.check:run',
        ],
    },
    'install_requires': (
        'abstracts>=0.0.12',
        'aio.core>=0.9.1',
        'aio.run.checker>=0.5.7',
        'envoy.base.utils>=0.3.9',
        'flake8>=5.0.4',
        'packaging>=23.0',
        'pyyaml',
        'yamllint',
        'yapf',
    ),
    'license': 'Apache Software License 2.0',
    'long_description': """
envoy.code.check
================

Code checker used in Envoy proxy's CI
""",
    'maintainer': 'Ryan Northey',
    'maintainer_email': 'ryan@synca.io',
    'name': 'envoy.code.check',
    'namespace_packages': (
    ),
    'package_data': {
        'envoy.code.check': (
            'py.typed',
        ),
    },
    'packages': (
        'envoy.code.check',
        'envoy.code.check.abstract',
    ),
    'url': 'https://github.com/envoyproxy/pytooling/tree/main/envoy.code.check',
    'version': '0.3.6',
})




import asyncio
import pathlib
from concurrent import futures
from typing import (
    AsyncIterator, Dict, Iterator, List, Optional, Set, Tuple, Type)

from packaging import version as _version

import abstracts

from aio.core import directory as _directory, event

from envoy.base import utils
from envoy.code.check import typing


class ICodeCheck(metaclass=abstracts.Interface):

    def __init__(
            self,
            directory: _directory.ADirectory,
            fix: bool = False,
            binaries: Optional[Dict[str, str]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        raise NotImplementedError


class IFileCodeCheck(ICodeCheck, metaclass=abstracts.Interface):

    @property  # type:ignore
    @abstracts.interfacemethod
    async def files(self) -> Set[str]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def problem_files(self) -> Dict[str, List[str]]:
        raise NotImplementedError


class IProjectCodeCheck(ICodeCheck, metaclass=abstracts.Interface):
    project: utils.interface.IProject

    def __init__(
            self,
            project: utils.interface.IProject,
            **kwargs) -> None:
        raise NotImplementedError


class IExtensionsCheck(metaclass=abstracts.Interface):

    def __init__(self, directory: _directory.ADirectory, **kwargs) -> None:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def all_fuzzed(self) -> bool:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def extensions_schema(self) -> "typing.ExtensionsSchemaDict":
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def fuzz_test_path(self) -> pathlib.Path:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def metadata(self) -> "typing.ExtensionsMetadataDict":
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def metadata_errors(self) -> Dict[str, Tuple[str, ...]]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def registration_errors(self) -> List[str]:
        raise NotImplementedError


class IFlake8Check(IFileCodeCheck, metaclass=abstracts.Interface):
    pass


class IGlintCheck(IFileCodeCheck, metaclass=abstracts.Interface):
    pass


class IShellcheckCheck(IFileCodeCheck, metaclass=abstracts.Interface):
    pass


class IYamllintCheck(IFileCodeCheck, metaclass=abstracts.Interface):
    pass


class IYapfCheck(IFileCodeCheck, metaclass=abstracts.Interface):
    pass


class IRuntimeGuardsCheck(IProjectCodeCheck, metaclass=abstracts.Interface):

    @property  # type:ignore
    @abstracts.interfacemethod
    async def missing(self) -> Set[str]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    async def status(self) -> AsyncIterator[Tuple[str, Optional[bool]]]:
        raise NotImplementedError


class IChangelogChangesChecker(metaclass=abstracts.Interface):

    @abstracts.interfacemethod
    def check_sections(
            self,
            version: _version.Version,
            sections: utils.typing.ChangelogChangeSectionsDict) -> (
                Tuple[str, ...]):
        raise NotImplementedError


class IChangelogStatus(metaclass=abstracts.Interface):
    pass


class IChangelogCheck(IProjectCodeCheck, metaclass=abstracts.Interface):

    def __init__(self, project: utils.interface.IProject, **kwargs) -> None:
        raise NotImplementedError

    @abstracts.interfacemethod
    def __iter__(self) -> Iterator[IChangelogStatus]:
        while False:
            yield NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def changes_checker_class(
            self) -> Type[IChangelogChangesChecker]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def changes_checker(self) -> IChangelogChangesChecker:
        raise NotImplementedError


class IRSTCheck(metaclass=abstracts.Interface):

    @abstracts.interfacemethod
    def __call__(self, text: str) -> Optional[str]:
        raise NotImplementedError


class ICodeChecker(event.IReactive, metaclass=abstracts.Interface):
    pass




import io
import logging
import os
import pathlib
from functools import cached_property, lru_cache
from typing import Dict, List, Set, Tuple

from flake8.main.application import Application  # type:ignore
from flake8 import (  # type:ignore
    utils as flake8_utils,
    checker as flake8_checker)

import abstracts

from aio.core.functional import async_property
from aio.core.directory.utils import directory_context
from aio.run import checker

from envoy.code.check import abstract, interface, typing


FLAKE8_CONFIG = '.flake8'


# Workaround for https://github.com/PyCQA/flake8/issues/1390
# logging.getLogger("flake8.options.manager").setLevel(logging.ERROR)
logger = logging.getLogger(__name__)


class Flake8Application(Application):
    """Subclassed flake8.Application to capture output."""

    @cached_property
    def output_fd(self) -> io.StringIO:
        return io.StringIO()

    def make_formatter(self) -> None:
        # ~Hacky workaround to capture flake8 output
        super().make_formatter()
        self.formatter.output_fd = self.output_fd
        self._formatter_stop = self.formatter.stop
        self.formatter.stop = self._stop

    def _stop(self) -> None:
        self.output_fd.seek(0)
        self._results: List[str] = [
            x
            for x
            in self.output_fd.read().strip().split("\n")
            if x]
        self._formatter_stop()


class Flake8App:
    """Wrapper around `flake8.main.application.Application`

    Provides optimized file discovery using app's lookup tools.
    """

    def __init__(self, path: str, args: Tuple[str, ...]) -> None:
        self.path = path
        self.args = args

    @cached_property
    def app(self) -> Flake8Application:
        """Flake8 Application."""
        flake8_app = Flake8Application()
        flake8_app.initialize(self.args)
        return flake8_app

    @property
    def exclude(self) -> List[str]:
        return self.manager.options.exclude

    @property
    def manager(self) -> flake8_checker.Manager:
        """Flake8 file checker manager."""
        return self.app.file_checker_manager

    def include_file(self, path: str) -> bool:
        """Include file according to flake8 config."""
        path = os.path.join(self.path, path)
        return (
            self._filename_matches(path)
            and self._include_directory(os.path.dirname(path))
            and not self._is_excluded(path))

    def include_files(self, files: Set[str]) -> Set[str]:
        """Figure out whether to include a file for checking."""
        return set(
            path
            for path
            in files
            if self.include_file(os.path.join(self.path, path)))

    def run_checks(self, paths: Set[str]) -> List[str]:
        """Run flake8 checks."""
        with directory_context(self.path):
            self.app.file_checker_manager.start(paths)
            self.app.report()
        return self.app._results

    @cached_property
    def _excluded_paths(self) -> Set[str]:
        return set()

    def _filename_matches(self, path: str) -> bool:
        return flake8_utils.fnmatch(
            path,
            self.app.options.filename)

    @lru_cache
    def _include_directory(self, path: str) -> bool:
        while True:
            if path == self.path:
                return True
            if not self._include_path(path):
                return False
            path = os.path.dirname(path)

    @lru_cache
    def _include_path(self, path: str) -> bool:
        exclude = (
           any(path.startswith(x) for x in self._excluded_paths)
           or self._is_excluded(path))
        if exclude:
            self._excluded_paths.add(path)
        return not exclude

    def _is_excluded(self, path: str) -> bool:
        if not self.exclude:
            return False
        basename = os.path.basename(path)
        if flake8_utils.fnmatch(basename, self.exclude):
            logger.debug(f'"{basename}" has been excluded')
            return True

        absolute_path = os.path.abspath(path)
        match = flake8_utils.fnmatch(absolute_path, self.exclude)
        logger.debug(
            f'"{absolute_path}" has {"" if match else "not "}been excluded')
        return match


@abstracts.implementer(interface.IFlake8Check)
class AFlake8Check(abstract.AFileCodeCheck, metaclass=abstracts.Abstraction):
    """Flake8 check for a fileset."""

    @classmethod
    def check_flake8_files(
            cls,
            path: str,
            args: Tuple[str, ...],
            files: Set[str]) -> List[str]:
        """Flake8 checker."""
        return Flake8App(
            path,
            args).run_checks(files)

    @classmethod
    def filter_flake8_files(
            cls,
            path: str,
            args: Tuple[str, ...],
            files: Set[str]) -> Set[str]:
        """Flake8 file discovery."""
        return Flake8App(
            path,
            args).include_files(files)

    @async_property
    async def checker_files(self) -> Set[str]:
        return await self.execute(
            self.filter_flake8_files,
            self.directory.absolute_path,
            self.flake8_args,
            await self.directory.files)

    @property
    def flake8_args(self) -> Tuple[str, ...]:
        """Flake configuration args."""
        return (
            "--config",
            str(self.flake8_config_path),
            str(self.directory.path))

    @property
    def flake8_config_path(self) -> pathlib.Path:
        """Path to flake8 configuration."""
        return self.directory.path.joinpath(FLAKE8_CONFIG)

    @async_property
    async def flake8_errors(self) -> List[str]:
        """Flake8 error list for check files."""
        # Important dont send an empty set to the flake8 checker,
        # as flake8 will check every file in path.
        return (
            await self.execute(
                self.check_flake8_files,
                self.directory.absolute_path,
                self.flake8_args,
                await self.files)
            if await self.files
            else [])

    @async_property(cache=True)
    async def problem_files(self) -> typing.ProblemDict:
        """Discovered flake8 errors."""
        return self.handle_errors(await self.flake8_errors)

    def handle_errors(self, errors: List[str]) -> typing.ProblemDict:
        """Turn flake8 error list -> `ProblemDict`."""
        flake8_errors: Dict[str, List[str]] = {}
        for error in errors:
            path, message = self._parse_error(error)
            flake8_errors[path] = flake8_errors.get(path, [])
            flake8_errors[path].append(message)
        return {
            p: checker.Problems(errors=errors)
            for p, errors
            in flake8_errors.items()}

    def _parse_error(self, error: str) -> Tuple[str, str]:
        path = error.split(":")[0]
        return (
            path,
            f"{path}: {error.split(':', 1)[1]}")




import asyncio
import re
from functools import cached_property, partial
from typing import (
    Callable, Iterable, Iterator,
    Pattern, Set, Tuple)

import abstracts

from aio.core.dev import debug
from aio.core import directory
from aio.core.functional import async_property
from aio.run import checker

from envoy.base import utils
from envoy.code.check import abstract, interface, typing


NOGLINT_RE = (
    r"[\w\W/-]*\.go$",
    r"[\w\W/-]*\.patch$",
    r"^test/[\w/]*_corpus/[\w/]*",
    r"^tools/[\w/]*_corpus/[\w/]*",
    r"[\w/]*password_protected_password.txt$")


@abstracts.implementer(directory.IDirectoryContext)
class NewlineChecker(directory.ADirectoryContext):

    @debug.logging(
        log=__name__,
        show_cpu=True)
    def no_newlines(self, paths: Iterable[str]) -> Set[str]:
        """Check files for final newline."""
        with self.in_directory:
            return set(
                target
                for target
                in paths
                if not self._has_newline(target))

    def _has_newline(self, target) -> bool:
        return (
            utils.last_n_bytes_of(target)
            == b'\n')


@abstracts.implementer(interface.IGlintCheck)
class AGlintCheck(abstract.AFileCodeCheck, metaclass=abstracts.Abstraction):

    @classmethod
    def no_newlines(cls, path: str, *paths: str) -> Set[str]:
        """Check files for final newline."""
        return NewlineChecker(path).no_newlines(paths)

    @classmethod
    def filter_files(
            cls, files: Set[str],
            match: Callable) -> Set[str]:
        """Filter files for `glint` checking."""
        return set(
            path
            for path
            in files
            if not match(path))

    @async_property
    async def checker_files(self) -> Set[str]:
        return self.filter_files(
            await self.directory.files,
            self.noglint_re.match)

    @async_property
    async def files_with_mixed_tabs(self) -> Set[str]:
        """Files with mixed preceeding tabs and spaces."""
        return await self.directory.grep(
            ["-lP", r"^ "],
            target=await self.files_with_preceeding_tabs)

    @async_property
    async def files_with_preceeding_tabs(self) -> Set[str]:
        """Files with preceeding tabs."""
        return await self.directory.grep(
            ["-lP", r"^\t"],
            target=await self.files)

    @async_property
    async def files_with_no_newline(self) -> Set[str]:
        """Files with no final newline."""
        batched = self.execute_in_batches(
            partial(self.no_newlines, self.directory.path),
            *await self.files)
        no_newline = set()
        async for batch in batched:
            no_newline |= batch
        return no_newline

    @async_property
    async def files_with_trailing_whitespace(self) -> Set[str]:
        """Files with trailing whitespace."""
        return await self.directory.grep(
            ["-lE", "[[:blank:]]$"],
            target=await self.files)

    @cached_property
    def noglint_re(self) -> Pattern[str]:
        """Regex for matching files that should not be checked."""
        return re.compile(r"|".join(NOGLINT_RE))

    @async_property(cache=True)
    async def problem_files(self) -> typing.ProblemDict:
        return (
            await self._check_problems(
                await asyncio.gather(
                    self.files_with_no_newline,
                    self.files_with_mixed_tabs,
                    self.files_with_trailing_whitespace))
            if await self.files
            else {})

    async def _check_problems(
            self,
            problems: Tuple[
                Set[str], Set[str], Set[str]]) -> typing.ProblemDict:
        return {
            path: checker.Problems(
                errors=list(self._check_path(path, *problems)))
            for path
            in (problems[0] | problems[1] | problems[2])}

    def _check_path(
            self,
            path: str,
            no_newline: Set[str],
            mixed_tabs: Set[str],
            trailing_whitespace: Set[str]) -> Iterator[str]:
        if path in no_newline:
            yield f"Missing final newline: {path}"
        if path in mixed_tabs:
            yield f"Mixed preceeding tabs and whitespace: {path}"
        if path in trailing_whitespace:
            yield f"Trailing whitespace: {path}"
